r"""
.. currentmodule: deeptime.basis

.. autosummary::
    :toctree: generated/
    :template: class_nomodule.rst

    Observable
    Concatenation

    Identity
    Monomials
"""

from ._base import Observable, Concatenation
from ._monomials import Identity, Monomials
